#include <intrin.h>
#include "aes.h"

static unsigned int t_fn[4][256];
static unsigned int t_fl[4][256];
static unsigned int t_in[4][256];
static unsigned int t_il[4][256];

__attribute__((optimize("-Os")))
void aes_init(aes_key *key, void *key_init)
{
	unsigned int i, w, b;
	unsigned int pow[510], log[256], t_rc[7], t_im[256];
	unsigned int *ss = key->ks_enc;

	i = 0; w = 1;
	do {
		pow[i] = pow[i + 255] = w;
		log[w] = i++;
		w ^= (w << 1) ^ ((w >> 7) * 0x011B);
	} while (w != 1);

	for (i = 0; i < 7; i++) {
		t_rc[i] = w;
		w = pow[log[w] + 0x19];
	}

	for (i = 0; i < 256; i++) {
		b = i ? pow[255 - log[i]] : 0;
		b ^= (b << 1) ^ (b << 2) ^ (b << 3) ^ (b << 4);
		b = ((b ^ (b >> 8)) & 0xff) ^ 0x63;

		t_fn[0][i] = w = b ? (pow[log[b] + 0x01] << 24) | pow[log[b] + 0x19] | b * 0x00010100 : 0;
		t_fn[1][i] = _rotl(w, 8);
		t_fn[2][i] = _rotl(w, 16);
		t_fn[3][i] = _rotl(w, 24);

		t_fl[0][i] = b;
		t_fl[1][i] = b << 8;
		t_fl[2][i] = b << 16;
		t_fl[3][i] = b << 24;

		w = (i << 1) ^ (i << 3) ^ (i << 6);
		w = ((w ^ (w >> 8)) & 0xff) ^ 0x05;
		b = w ? pow[255 - log[w]] : 0;

		t_in[0][i] = t_im[b] = w = b ? (pow[log[b] + 0x68] << 24) | (pow[log[b] + 0xee] << 16) | (pow[log[b] + 0xc7] << 8) | pow[log[b] + 0xdf] : 0;
		t_in[1][i] = _rotl(w, 8);
		t_in[2][i] = _rotl(w, 16);
		t_in[3][i] = _rotl(w, 24);

		t_il[0][i] = b;
		t_il[1][i] = b << 8;
		t_il[2][i] = b << 16;
		t_il[3][i] = b << 24;
	}

	for (i = 0; i < 8; i++)
		key->ks_enc[i] = key->ks_dec[56 + i - 2 * (i & 4)] = ((unsigned int*)key_init)[i];

	for (i = 0; i < 52; i++)
		key->ks_enc[i + 8] = key->ks_dec[48 + (i & 3) - (i & ~7) - (i & 4)] =
			ss[i & 7] ^= (i & 7) == 0 ? t_fl[0][(ss[7] >> 8) & 0xff] ^ t_fl[1][(ss[7] >> 16) & 0xff] ^ t_fl[2][ss[7] >> 24] ^ t_fl[3][ss[7] & 0xff] ^ t_rc[i >> 3] :
						 (i & 7) == 4 ? t_fl[0][ss[3] & 0xff] ^ t_fl[1][(ss[3] >> 8) & 0xff] ^ t_fl[2][(ss[3] >> 16) & 0xff] ^ t_fl[3][ss[3] >> 24] :
						 ss[(i & 7) - 1];

	memcpy(key->ks_enc, key_init, 32);

	for (i = 4; i < 14 * 4; i++)
		key->ks_dec[i] = t_im[key->ks_dec[i] & 0xff] ^ _rotl(t_im[(key->ks_dec[i] >> 8) & 0xff], 8) ^ _rotl(t_im[(key->ks_dec[i] >> 16) & 0xff], 16) ^ _rotl(t_im[key->ks_dec[i] >> 24], 24);
}

#ifndef _WIN64
__attribute__((optimize("-O3")))
#else
__attribute__((optimize("-O3")))
#endif
void aes_encrypt_sw(aes_key *key, void *buffer)
{
	unsigned int b0[4], b1[4], rnd;
	unsigned int *kp = key->ks_enc;

	b0[0] = ((unsigned int*)buffer)[0] ^ kp[0];
	b0[1] = ((unsigned int*)buffer)[1] ^ kp[1];
	b0[2] = ((unsigned int*)buffer)[2] ^ kp[2];
	b0[3] = ((unsigned int*)buffer)[3] ^ kp[3];

	rnd = 13;
	do {
		kp += 4;
		b1[0] = kp[0] ^ t_fn[0][b0[0] & 0xff] ^ t_fn[1][(b0[1] >> 8) & 0xff] ^ t_fn[2][(b0[2] >> 16) & 0xff] ^ t_fn[3][b0[3] >> 24];
		b1[1] = kp[1] ^ t_fn[0][b0[1] & 0xff] ^ t_fn[1][(b0[2] >> 8) & 0xff] ^ t_fn[2][(b0[3] >> 16) & 0xff] ^ t_fn[3][b0[0] >> 24];
		b1[2] = kp[2] ^ t_fn[0][b0[2] & 0xff] ^ t_fn[1][(b0[3] >> 8) & 0xff] ^ t_fn[2][(b0[0] >> 16) & 0xff] ^ t_fn[3][b0[1] >> 24];
		b1[3] = kp[3] ^ t_fn[0][b0[3] & 0xff] ^ t_fn[1][(b0[0] >> 8) & 0xff] ^ t_fn[2][(b0[1] >> 16) & 0xff] ^ t_fn[3][b0[2] >> 24];
		b0[0] = b1[0]; b0[1] = b1[1]; b0[2] = b1[2]; b0[3] = b1[3];
	} while (--rnd);
	b0[0] = kp[4] ^ t_fl[0][b1[0] & 0xff] ^ t_fl[1][(b1[1] >> 8) & 0xff] ^ t_fl[2][(b1[2] >> 16) & 0xff] ^ t_fl[3][b1[3] >> 24];
	b0[1] = kp[5] ^ t_fl[0][b1[1] & 0xff] ^ t_fl[1][(b1[2] >> 8) & 0xff] ^ t_fl[2][(b1[3] >> 16) & 0xff] ^ t_fl[3][b1[0] >> 24];
	b0[2] = kp[6] ^ t_fl[0][b1[2] & 0xff] ^ t_fl[1][(b1[3] >> 8) & 0xff] ^ t_fl[2][(b1[0] >> 16) & 0xff] ^ t_fl[3][b1[1] >> 24];
	b0[3] = kp[7] ^ t_fl[0][b1[3] & 0xff] ^ t_fl[1][(b1[0] >> 8) & 0xff] ^ t_fl[2][(b1[1] >> 16) & 0xff] ^ t_fl[3][b1[2] >> 24];

	((unsigned int*)buffer)[0] = b0[0];
	((unsigned int*)buffer)[1] = b0[1];
	((unsigned int*)buffer)[2] = b0[2];
	((unsigned int*)buffer)[3] = b0[3];
}

void aes_encrypt_hw(aes_key *key, void *buffer)
{
	__m128i tmp;
	unsigned int rnd;

	tmp = *(__m128i*)buffer ^ ((__m128i*)key->ks_enc)[0];
	for (rnd = 1; rnd < 14; rnd++)
		tmp = _mm_aesenc_si128(tmp, ((__m128i*)key->ks_enc)[rnd]);
	*(__m128i*)buffer = _mm_aesenclast_si128(tmp, ((__m128i*)key->ks_enc)[14]);
}


#ifndef _WIN64
__attribute__((optimize("-O1")))
#else
__attribute__((optimize("-O3")))
#endif
void aes_decrypt_sw(aes_key *key, void *buffer)
{
	unsigned int b0[4], b1[4], rnd;
	unsigned int *kp = key->ks_dec;

	b0[0] = ((unsigned int*)buffer)[0] ^ kp[0];
	b0[1] = ((unsigned int*)buffer)[1] ^ kp[1];
	b0[2] = ((unsigned int*)buffer)[2] ^ kp[2];
	b0[3] = ((unsigned int*)buffer)[3] ^ kp[3];

	rnd = 13;
	do {
		kp += 4;
		b1[0] = kp[0] ^ t_in[0][b0[0] & 0xff] ^ t_in[1][(b0[3] >> 8) & 0xff] ^ t_in[2][(b0[2] >> 16) & 0xff] ^ t_in[3][b0[1] >> 24];
		b1[1] = kp[1] ^ t_in[0][b0[1] & 0xff] ^ t_in[1][(b0[0] >> 8) & 0xff] ^ t_in[2][(b0[3] >> 16) & 0xff] ^ t_in[3][b0[2] >> 24];
		b1[2] = kp[2] ^ t_in[0][b0[2] & 0xff] ^ t_in[1][(b0[1] >> 8) & 0xff] ^ t_in[2][(b0[0] >> 16) & 0xff] ^ t_in[3][b0[3] >> 24];
		b1[3] = kp[3] ^ t_in[0][b0[3] & 0xff] ^ t_in[1][(b0[2] >> 8) & 0xff] ^ t_in[2][(b0[1] >> 16) & 0xff] ^ t_in[3][b0[0] >> 24];
		b0[0] = b1[0]; b0[1] = b1[1]; b0[2] = b1[2]; b0[3] = b1[3];
	} while (--rnd);
	b0[0] = kp[4] ^ t_il[0][b1[0] & 0xff] ^ t_il[1][(b1[3] >> 8) & 0xff] ^ t_il[2][(b1[2] >> 16) & 0xff] ^ t_il[3][b1[1] >> 24];
	b0[1] = kp[5] ^ t_il[0][b1[1] & 0xff] ^ t_il[1][(b1[0] >> 8) & 0xff] ^ t_il[2][(b1[3] >> 16) & 0xff] ^ t_il[3][b1[2] >> 24];
	b0[2] = kp[6] ^ t_il[0][b1[2] & 0xff] ^ t_il[1][(b1[1] >> 8) & 0xff] ^ t_il[2][(b1[0] >> 16) & 0xff] ^ t_il[3][b1[3] >> 24];
	b0[3] = kp[7] ^ t_il[0][b1[3] & 0xff] ^ t_il[1][(b1[2] >> 8) & 0xff] ^ t_il[2][(b1[1] >> 16) & 0xff] ^ t_il[3][b1[0] >> 24];

	((unsigned int*)buffer)[0] = b0[0];
	((unsigned int*)buffer)[1] = b0[1];
	((unsigned int*)buffer)[2] = b0[2];
	((unsigned int*)buffer)[3] = b0[3];
}

void aes_decrypt_hw(aes_key *key, void *buffer)
{
	__m128i tmp;
	unsigned int rnd;

	tmp = *(__m128i*)buffer ^ ((__m128i*)key->ks_dec)[0];
	for (rnd = 1; rnd < 14; rnd++)
		tmp = _mm_aesdec_si128(tmp, ((__m128i*)key->ks_dec)[rnd]);
	*(__m128i*)buffer = _mm_aesdeclast_si128(tmp, ((__m128i*)key->ks_dec)[14]);
}
